#! /usr/bin/env python
# -*- coding: utf-8 -*-

from saqc.flagger.baseflagger import BaseFlagger
from saqc.flagger.categoricalflagger import CategoricalFlagger
from saqc.flagger.simpleflagger import SimpleFlagger
from saqc.flagger.dmpflagger import DmpFlagger
from saqc.flagger.continuousflagger import ContinuousFlagger
