#! /usr/bin/env python
# -*- coding: utf-8 -*-

# imports needed to make the functions register themself
from .register import register
from .functions import *
from .break_detection import *
from .constants_detection import *
from .soil_moisture_tests import *
from .spike_detection import *
from .harm_functions import *
from .machine_learning import *
